/*

Converts an Atari PI1/PI2/PI3 Degas Elite picture file into raw data + palette

(c) 2001 - Laurent de Soras (laurent@ohmforce.com)

Raw data is a 256 indexed color file
Palette is .pal format for JASC Paint Shop Pro, 256 colors
PI1 = 320x200, 16 colors
PI2 = 640x200, 4 colors
PI3 = 640x400, 2 colors

*/



#if defined (_MSC_VER)
	#pragma warning (4 : 4786)
#endif



#include	<fstream>
#include	<iostream>
#include	<string>
#include	<vector>

#include	<cassert>



typedef	unsigned short	UInt16;
typedef	unsigned char	UByte;



// -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -



class DegasEliteFile
{
public:

	enum {			DATA_SIZE		= 320L * 200	};
	enum {			MAX_NBR_COLORS	= 16	};
	enum Type
	{
		Type_16_COL	= 0,
		Type_4_COL,
		Type_2_COL,
		Type_COMPRESSED = 0x8000,
		Type_NOT_LOADED = 0xFFFF
	};

						DegasEliteFile ();

	int				load_file (const std::string &filename);

	UInt16			_type;
	UInt16			_palette [MAX_NBR_COLORS];
	UByte				_data [DATA_SIZE];
};



DegasEliteFile::DegasEliteFile ()
:	_type (Type_NOT_LOADED)
{
}



int	DegasEliteFile::load_file (const std::string &filename)
{
	int				ret_val = 0;

	std::ifstream	ifs (filename.c_str (), std::ios::binary);

	std::streambuf *	isb_ptr = ifs.rdbuf ();
	if (isb_ptr == 0)
	{
		ret_val = -1;
	}

	// Load Info
	if (ret_val == 0)
	{
		_type = isb_ptr->sbumpc () * 0x100;
		_type += isb_ptr->sbumpc ();
		if (! ifs)
		{
			ret_val = -1;
		}
	}

	// Load Palette
	if (ret_val == 0)
	{
		for (int col = 0; col < MAX_NBR_COLORS && ret_val == 0; ++col)
		{
			_palette [col] = isb_ptr->sbumpc () * 0x100;
			_palette [col] += isb_ptr->sbumpc ();
			if (! ifs)
			{
				ret_val = -1;
			}
		}
	}

	// Load data
	if (ret_val == 0)
	{
		isb_ptr->sgetn (reinterpret_cast <char *> (_data),	DATA_SIZE);
		if (! ifs)
		{
			ret_val = -1;
		}
	}

	ifs.close ();

	return (ret_val);
}



// -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -



class SimplePicture
{
public:

						SimplePicture ();
	int				convert_from_degas_elite (const DegasEliteFile &pic);
	int				save_palette (const std::string &filename);
	int				save_data (const std::string &filename);

private:

	typedef	std::vector <UByte>	Row;
	struct	Color
	{
		int			_r;
		int			_g;
		int			_b;
	};

	int				set_size (int width, int height);
	int				convert_pix_from_degas_elite (const DegasEliteFile &pic, int bits_per_pixel);
	int				convert_pal_from_degas_elite (const DegasEliteFile &pic, int nbr_colors);

	int				_width;
	int				_height;
	std::vector <Row>
						_pixels;
	std::vector <Color>
						_palette;
};



SimplePicture::SimplePicture ()
:	_width (0)
,	_height (0)
{
}



int	SimplePicture::convert_from_degas_elite (const DegasEliteFile &pic)
{
	int				ret_val = 0;

	// Set size and bit depth
	int				bits_per_pixel;
	switch (pic._type)
	{
	case	DegasEliteFile::Type_16_COL:
		bits_per_pixel = 4;
		ret_val = set_size (320, 200);
		break;

	case	DegasEliteFile::Type_4_COL:
		bits_per_pixel = 2;
		ret_val = set_size (640, 200);
		break;

	case	DegasEliteFile::Type_2_COL:
		bits_per_pixel = 1;
		ret_val = set_size (640, 400);
		break;

	default:
		ret_val = -1;
		break;
	}

	// Convert pixel data
	if (ret_val == 0)
	{
		ret_val = convert_pix_from_degas_elite (pic, bits_per_pixel);
	}

	// Convert palette
	if (ret_val == 0)
	{
		ret_val = convert_pal_from_degas_elite (pic, 1 << bits_per_pixel);
	}

	return (ret_val);
}



int	SimplePicture::save_palette (const std::string &filename)
{
	int				ret_val = 0;

	std::ofstream	ofs (filename.c_str ());

	const long		max_nbr_colors = 256;
	ofs << "JASC-PAL\n0100\n" << max_nbr_colors << '\n';
	int				col;

	const int		palette_size = _palette.size ();
	const int		limit =   (palette_size < max_nbr_colors)
						        ? palette_size
						        : max_nbr_colors;
	for (col = 0; col < limit && ret_val == 0; ++col)
	{
		ofs << _palette [col]._r << ' ';
		ofs << _palette [col]._g << ' ';
		ofs << _palette [col]._b << '\n';
		if (! ofs)
		{
			ret_val = -1;
		}
	}

	for (col = col ; col < max_nbr_colors && ret_val == 0; ++col)
	{
		ofs << "0 0 0\n";
		if (! ofs)
		{
			ret_val = -1;
		}
	}

	ofs.close ();

	return (ret_val);
}



int	SimplePicture::save_data (const std::string &filename)
{
	int				ret_val = 0;

	std::ofstream	ofs (filename.c_str (), std::ios::binary);
	std::streambuf *	osb_ptr = ofs.rdbuf ();
	if (osb_ptr == 0)
	{
		ret_val = -1;
	}

	for (int line = 0; line < _height && ret_val == 0; ++line)
	{
		if (osb_ptr->sputn (reinterpret_cast <const char *> (&_pixels [line] [0]), _width) != _width)
		{
			ret_val = -1;
		}
	}

	ofs.close ();

	return (ret_val);
}



int	SimplePicture::set_size (int width, int height)
{
	assert (width >= 0);
	assert (height >= 0);

	_width = width;
	_height = height;

	_pixels.resize (_height);
	for (int y = 0; y < _height; ++y)
	{
		_pixels [y].resize (_width);
	}

	return (0);
}



int	SimplePicture::convert_pix_from_degas_elite (const DegasEliteFile &pic, int bits_per_pixel)
{
	const int		pixels_per_byte = 8 / bits_per_pixel;
	const long		line_size = _width / pixels_per_byte;
	for (int y = 0; y < _height; ++y)
	{
		for (int x = 0; x < _width; x += 16)
		{
			const long		block_pos = y * line_size + x / pixels_per_byte;
			for (int x2 = 0; x2 < 16; ++x2)
			{
				const int		shift = 15 - x2;
				int				color = 0;
				for (int plane = 0; plane < bits_per_pixel; ++plane)
				{
					const long		plane_pos = block_pos + plane * 2;
					const int		data =   pic._data [plane_pos    ] * 0x100
										       + pic._data [plane_pos + 1];
					color += ((data >> shift) & 1) << plane;
				}
				_pixels [y] [x + x2] = color;
			}
		}
	}

	return (0);
}



int	SimplePicture::convert_pal_from_degas_elite (const DegasEliteFile &pic, int nbr_colors)
{
	assert (nbr_colors > 0);

	_palette.resize (nbr_colors);
	for (int col = 0; col < nbr_colors; ++col)
	{
		const int		r =   ((pic._palette [col] >> 7) & 0xE)
							    + ((pic._palette [col] >> 11) & 0x1);
		_palette [col]._r = r * 0x11;

		const int		g =   ((pic._palette [col] >> 3) & 0xE)
							    + ((pic._palette [col] >> 7) & 0x1);
		_palette [col]._g = g * 0x11;

		const int		b =   ((pic._palette [col] << 1) & 0xE)
							    + ((pic._palette [col] >> 3) & 0x1);
		_palette [col]._b = b * 0x11;
	}

	return (0);
}



// -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -



int	convert_degas_elite (const std::string &new_filename, const std::string &old_filename);



int	main (int argc, char *argv [])
{
	int				ret_val = 0;

	if (argc != 2)
	{
		std::cout << "Convert Atari Degas Elite files (.PI1, .PI2 and .PI3) to raw format.\n\n";
		std::cout << "Usage :\n" << argv [0] << " picture_filename\n\n";
		ret_val = -1;
	}

	else
	{
		const std::string	old_filename (argv [1]);
		const long		dot_pos = old_filename.find_last_of ('.');
		std::string	new_filename = old_filename.substr (0, dot_pos);

		ret_val = convert_degas_elite (new_filename, old_filename);
	}

	if (ret_val != 0)
	{
		std::cout << "*** Error.\n";
	}

	return (ret_val);
}



int	convert_degas_elite (const std::string &new_filename, const std::string &old_filename)
{
	int				ret_val = 0;

	SimplePicture		picture;
	DegasEliteFile		degas_file;

	// Load file
	ret_val = degas_file.load_file (old_filename);
	if (ret_val == 0)
	{
		ret_val = picture.convert_from_degas_elite (degas_file);
	}

	// Write data
	if (ret_val == 0)
	{
		const std::string	data_filename = new_filename + ".raw";
		ret_val = picture.save_data (data_filename);
	}

	// Write palette
	if (ret_val == 0)
	{
		const std::string	palette_filename = new_filename + ".pal";
		ret_val = picture.save_palette (palette_filename);
	}

	return (ret_val);
}
